/*
** ttree.c - test the mime_invariant checks of mime.c
*/
/*
** Part of MIMEC: see COPYING
** Copyright (c) 1999-2004 Alessandro Vesely
** All rights reserved.
**
** compile: gcc -W -Wall -O0 -g -o ttree -DTEST_SEEK ttree.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <limits.h>


static int 
recurse_level(unsigned not_discrete, unsigned last, size_t i,
	size_t const target, int level, int *result)
{
	assert(result);
	assert((int)i != -1);
	assert(i <= target);
	
	for (; i < target; ++i)
	{
		unsigned const m = 1 << i;
		if (m & not_discrete)
		{
			i = recurse_level(not_discrete, last, i + 1,
				target, level + 1, result);
			if ((int)i == -1)
				return -1;
		}
		if (m & last)
			return i;
	}
	
	assert(i == target);
	assert(level >= 0);
	
	*result = level;
	return -1;
}

static int get_level(unsigned not_discrete, unsigned last, size_t target)
{
	int result = -1;
	int i;
		
	i = recurse_level(not_discrete, last, 0, target, 0, &result);

	assert(i == -1);
	assert(result >= 0);
	return result;
}

static int seek_last_child(unsigned not_discrete, unsigned last, size_t i);

static int seek_last_sibling(unsigned not_discrete, unsigned last, size_t i)
/*
** if keeplevel != 0 return index of last sibling
*/
{
	unsigned m;
	
	while (((m = 1 << i) & last) == 0)
	{
		if (m & not_discrete)
			i = seek_last_child(not_discrete, last, i);
		++i;
	}

	return i;
}

static int seek_last_child(unsigned not_discrete, unsigned last, size_t i)
/*
** if keeplevel == 0 return index of last element in subtree,
*/
{
	unsigned m = 1 << i;
	
	if (m & not_discrete)
		i = seek_last_child(not_discrete, last,
			seek_last_sibling(not_discrete, last, i + 1));

	return i;
}




int reach_last_node(unsigned not_discrete, unsigned last, int i, int bc)
/*
** Definition of tree.
**
** Reach the last node (highest index) among the given node i, any sibling
** of i and any children thereof; when started from node 0 this reaches
** the absolutely last node (bc - 1) recursively. If this can be done then
** the pair (not_discrete, last) represents a tree.
*/
{
	if (i >= bc)
		return -1;
		
	for (; i < bc; ++i)
	{
		unsigned const  m = 1 << i;
		if (not_discrete & m)
		/*
		** if current node (i) has children, i+1 is the first one
		*/
		{
			i = reach_last_node(not_discrete, last, i + 1, bc);
			if (i < 0)
				return -1;
		}
		/*
		** if current node N is the last sibling, i is the last node:
		** obvious if it had no children; otherwise i is the last
		** node of the first child C: since N has no siblings, i is
		** also the last node of N.
		*/
		if (last & m)
			return i;
	}
	
	assert(0);
	return -1;
}

int passed_balance;
int mime_invariant(unsigned not_discrete, unsigned last, int bc)
{
	int balance, i, rtc;

	/*
	** the root node, if any, must have no siblings
	*/
	if (bc > 0 && (last & 1) == 0)
		return 0;		

	/*
	** among one node's children, exactly one must be the last
	** (these guarantees return from next recursive check)
	*/
	balance = 1; // the root node is the last, but is noone's child
	for (i = 0; i < bc; ++i)
	{
		unsigned const m = 1 << i;
		if (m & not_discrete)
			++balance;
		if (m & last)
		{
#if 1		
			--balance;
#else
// could save some cpu time with more stringent tests, e.g.
			if (--balance < 0)
				return 0;
#endif
		}
	}

	if (balance != 0)
		return 0;
	++passed_balance;
	rtc = reach_last_node(not_discrete, last, 0, bc) + 1 == bc;

#if defined(TEST_SEEK)
	if (rtc)
	{
		int tt;
		if ((tt=seek_last_child(not_discrete, last, 0)) != bc - 1)
			fprintf(stderr, "fail for nd=%#x, ls=%#x, bc = %d -> %d\n",
				not_discrete, last, bc, tt);
		assert(seek_last_child(not_discrete, last, 0) == bc - 1);
		
		for (i = 0; i < bc; ++i)
		{
			int sibling = seek_last_sibling(not_discrete, last, i);
			assert(get_level(not_discrete, last, i) ==
				get_level(not_discrete, last, sibling));
		}
	}
#endif
	
	return rtc;	
}

typedef struct ntree
{
	int up, left, right, first, last, here;
} ntree;

int set_ntree(ntree *nt, int i, int up, unsigned nd, unsigned ls, int bc)
/*
** recursively fill the elements of an ntree.
** nt is the base, containing enough elements,
** i is the start index (0 for root),
** up is the parent node (-1 for root),
** nd, ls, bc are the bit flags and count for the bit representation
*/
{
	int left = -1;

	assert(mime_invariant(nd, ls, bc));
	
	for (;;)
	{
		unsigned const m = 1 << i;
		int here = i;

		nt[here].here = here;
		nt[here].up = up;
		nt[here].left = left;
		left = i;

		if (nd & m)
			i = set_ntree(nt, nt[here].first = i + 1, here, nd, ls, bc);
		else
			nt[here].first = nt[here].last = -1;
		
		if (ls & m)
		{
			nt[here].right = -1;
			if (up >= 0)
				nt[up].last = here;
			return i;
		}
		else
			nt[here].right = ++i;
	}
}

int get_depth(ntree *nt, int i, int depth)
{
	if (nt[i].up >= 0)
		return get_depth(nt, nt[i].up, depth + 1);
	return depth;
}

int child_disp(ntree *nt, int here)
{
	int i, size;
	
	if (nt[here].first < 0)
		return 1;
	
	size = 0;
	for (i = nt[here].first; i >= 0; i = nt[i].right)
		size += child_disp(nt, i);
	
	return size;
}

int child_depth(ntree *nt, int here)
{
	int i, depth;
	
	if (nt[here].first < 0)
		return 1;
	
	depth = 0;
	for (i = nt[here].first; i >= 0; i = nt[i].right)
	{
		int c = child_depth(nt, i);
		if (c > depth)
			depth = c;
	}
	
	return depth + 1;
}

void draw_node(ntree *nt, int here, int ndx_size, int line, int disp)
{
	int left = nt[here].left;
	int indent = disp * (ndx_size + 2);
	if (line == 1)
	{
		int link = '|';
		if (left >= 0)
		{
			link = '\\';
		}
		
		indent += ndx_size;
		printf("%*s%c ", indent, "", link);
	}
	else
	{
		int link = ' ';
		if (left >= 0)
		{
			link = '-';
		}

		printf("%*s%c%*.*d%c",
			indent, "", link, ndx_size, ndx_size, nt[here].here,
			nt[here].right >= 0 ? '-' : ' ');
	}
}

int recurse_draw_tree(ntree *nt, int here, int ndx_size,
	int depth, int line, int disp)
{
	int i;
	
	if (depth == get_depth(nt, here, 0))
	{
		draw_node(nt, here, ndx_size, line, disp);
		return child_disp(nt, here) - 1;
	}
	
	i = nt[here].first;
	if (i < 0)
		return disp + 1;

	for (; i >= 0; i = nt[i].right)
		disp = recurse_draw_tree(nt, i, ndx_size, depth, line, disp);
	
	return disp;
}

void draw_tree(unsigned nd, unsigned ls, int bc, int no_table)
{
	ntree *nt = (ntree*)malloc(bc * sizeof(ntree));
	int ndx_size, depth, disp, i, d;
	
	assert(mime_invariant(nd, ls, bc));
	
	if (nt == NULL)
		return;

	set_ntree(nt, 0, -1, nd, ls, bc);
#if defined(TEST_SEEK)
	for (i = 0; i < bc; ++i)
		assert(get_level(nd, ls, i) == get_depth(nt, i, 0));
#endif
	
	
	for (ndx_size = 0, i = bc; i; i /= 10)
		++ndx_size;
	
	depth = child_depth(nt, 0);

	disp = recurse_draw_tree(nt, 0, ndx_size, 0, 0, 0);
	printf("%*s[0]\n", disp * (ndx_size + 2) + 2, "");
	for (d = 1; d < depth; ++d)
	   for (i = 1; i >= 0; --i)
		{
			disp = recurse_draw_tree(nt, 0, ndx_size, d, i, 0);
			printf("%*s[%d]\n", disp * (ndx_size + 2) + 2, "", d);
		}
	
	if (no_table)
	{
		putchar('\n');
	}
	else
	{
		if (ndx_size < 5)
			ndx_size = 5;
		printf("\n%*s %*s %*s %*s %*s %*s\n",
			ndx_size, "here",
			ndx_size, "up",
			ndx_size, "left",
			ndx_size, "right",
			ndx_size, "first",
			ndx_size, "last");
		for (i = 0; i < bc; ++i)
			printf("%*d %*d %*d %*d %*d %*d [%d]\n",
			ndx_size, nt[i].here,
			ndx_size, nt[i].up,
			ndx_size, nt[i].left,
			ndx_size, nt[i].right,
			ndx_size, nt[i].first,
			ndx_size, nt[i].last, get_depth(nt, i, 0));
	}
	free(nt);
}

typedef struct possible_tree
{
	unsigned bit_count;
	unsigned nd, ls;
} possible_tree;

int bitcompare(unsigned a, unsigned b)
{
	unsigned m;
	for (m = 1; m != 0; m <<= 1)
		if ((m & a) != (m & b))
			return (m & a) ? -1 : 1;
	return 0;
}

int consecutive(unsigned a)
{
	unsigned m, mp = 1, count = 0;
	for (m = 1; m != 0; m <<= 1)
	{
		unsigned p = (m & a) != 0;
		unsigned pp = (mp & count) != 0;
		if (p != pp)
		{
			count |= p;
			pp <<= 1;
		}
	}
	return count;
}

int bitcompareconsecutive(unsigned a, unsigned b)
{
	unsigned ca = consecutive(a), cb = consecutive(b);
	unsigned bc = bitcompare(ca, cb);
	if (bc == 0)
		return bitcompare(a, b);
	else
		return bc;
}

int compare_possible_tree(void const *pa, void const *pb)
{
	possible_tree const * const a = (possible_tree const * const)pa;
	possible_tree const * const b = (possible_tree const * const)pb;
	
	if (a->bit_count == b->bit_count)
	{
		if (a->nd == b->nd)
		{
			if (a->ls == b->ls)
				return 0;
				
			return bitcompare(a->ls, b->ls);
		}
		
		return bitcompareconsecutive(a->nd, b->nd);
	}
	
	return a->bit_count < b->bit_count ? -1 : 1;
}

void report_tree_table(possible_tree *pt, size_t start, size_t count)
{
	unsigned m, j, i, bc = pt->bit_count;

	for (i = start; i < count; ++i)
		printf(" %c |", start == 0 ? i + 'A' : i % 10 + '0');
	putc('\n', stdout);
	
	for (i = start; i < count; ++i)
		fputs("---|", stdout);
	putc('\n', stdout);

	m = 1;
	for (j = 0; j < bc; ++j)
	{
		for (i = start; i < count; ++i)
			printf(" %c |", ((pt[i].nd & m) != 0) + '0');
		m <<= 1;
		if (j == 0)
			puts(" non discrete");
		else
			putc('\n', stdout);
	}
	
	for (i = start; i < count; ++i)
		fputs("---|", stdout);
	putc('\n', stdout);

	m = 1;
	for (j = 0; j < bc; ++j)
	{
		for (i = start; i < count; ++i)
			printf(" %c |", ((pt[i].ls & m) != 0) + '0');
		m <<= 1;
		if (j == 0)
			puts(" last");
		else
			putc('\n', stdout);
	}
	putc('\n', stdout);
}

int report_tree(possible_tree *pt, size_t lcount, int report)
{
	unsigned i, bc = pt->bit_count, count;
	int tt = -1;
	
	for (i = 0; i < lcount; ++i)
		assert(bc == pt[i].bit_count);
	printf("%d tree(s) with %d nodes\n", lcount, bc);
	
	if (lcount > 15)
		count = 10;
	else
		count = lcount;
	report_tree_table(pt, 0, count);

	while (report)
	{
		unsigned char ttb[8];
		char *s;
		
		fputs("draw which tree? ", stdout);
		if (count != lcount)
		{
			char *end;
			unsigned long ul;
			printf("(1-%d [%d]) ", lcount, tt >= 0 ? tt + 2 : 1);
			fflush(stdout);
			s = fgets(ttb, sizeof ttb, stdin);
			if (s != NULL)
			{
				if ((ul = strtoul(ttb, &end, 0)) < INT_MAX &&
					*end == '\n' && ul > 0 && ul <= lcount)
				{
						tt = (int)ul;
						count = tt > 5 ? tt - 5 : 0;
						report_tree_table(pt, count - 5, count + 5);
				}
				else if (ttb[0] == '\n' || ttb[0] == '+')
				{
					if (tt >= 0)
						tt += 2;
					else
						tt = 1;
					ttb[0] = '+';
				}
				else
					return -1;
				
				if ((size_t)tt <= lcount)
					--tt;
			}
		}
		else
		{
			printf("(A-%c [none]) ", count - 1 + 'A');
			fflush(stdout);
			s = fgets(ttb, sizeof ttb, stdin);
			putchar('\n');
			if (s != NULL &&
				ttb[1] == '\n' &&
				ttb[0] >= 'A' &&
				ttb[0] <= count - 1 + 'A')
					tt = ttb[0] - 'A';
				else
					return -1;
		}
		if (tt >= 0 && (size_t)tt < lcount)
			draw_tree(pt[tt].nd, pt[tt].ls, bc, ttb[0] == '+');
		else
			return 1;
	}
	
	return 0;
}

int main(int argc, char *argv[])
{
	int max_p = 0, max_bc = 0, pcount = 0;
	int report_only = 0;
	int bc, nd, ls;
	int i, j;
	possible_tree *pt;
	
	if (argc > 3)
		max_p = atoi(argv[3]);
	
	if (max_p <= 0)
		max_p = 128;
	
	if (argc > 2)
		report_only = 1; /* atoi(argv[2]); */

	if (argc > 1)
		max_bc = atoi(argv[1]);
	
	/* don't try an argument larger than 6
	** (alphabet and line length limits reached)
	*/
	if (max_bc <= 0 || (size_t)max_bc > 8*sizeof(unsigned))
		max_bc = 5;
		
	pt = (possible_tree*)malloc(max_p * sizeof(possible_tree));
	if (pt == NULL)
		return 1;
	
	passed_balance = 0;
	for (bc = 0; bc < max_bc; ++bc)
	{
		int maxn = 1 << bc;
		for (nd = 0; nd < maxn; ++nd)
			for (ls = 0; ls < maxn; ++ls)
				if (mime_invariant(nd, ls, bc))
				{
					if (pcount < max_p)
					{
						pt[pcount].nd = nd;
						pt[pcount].ls = ls;
						pt[pcount].bit_count = bc;
					}
					++pcount;
				}
	}
	
	printf("%d/%d tree(s) found with less than %d nodes\n"
		"%d passed balance test\n\n",
		pcount, max_p, max_bc, passed_balance);

	if (max_bc <= 6 && pcount < max_p)
	{
		qsort(pt, pcount, sizeof pt[0], &compare_possible_tree);
		
		for (i = 0; i < pcount; i = j)
		{
			unsigned bc = pt[i].bit_count;
			for (j = i + 1; j < pcount; ++j)
				if (bc != pt[j].bit_count)
					break;
					
			if ((report_only == 0 || pcount == j) &&
				report_tree(pt + i, j - i, report_only) != 0)
					break;
		}
	}
	
	free(pt);
	return 0;
}
